"""
Small utils and tools to use with the Messagebus.
"""
import json
import logging


def create_echo_function(name):
    """Standard logging mechanism for Alice processes.

    This creats
    Arguments:
        name (str): Reference name of the process

    Returns:
        func: The echo function
    """
    log = logging.getLogger(name)

    def echo(message):
        try:
            msg = json.loads(message)
            msg_type = msg.get("type", "")
            # do not log tokens from registration messages
            if msg_type == "registration":
                msg["data"]["token"] = None
                message = json.dumps(msg)
        except Exception as exc:
            log.info("Error: %s", repr(exc), exc_info=True)

        # Listen for messages and echo them for logging
        log.info("BUS: %s", repr(message))
    return echo



"""
Send offers a simple entry point for scripts to send a single  message on
the alice messagebus.

A bash script can run

python -c "from alice_bus_client.send_func import send; \
send('speak', {'utterance': 'hello'})"
"""
from websocket import create_connection

from .client import MessageBusClient
from .message import Message


def send(message_to_send, data_to_send=None, config=None):
    """Send a single message over the websocket.

    Args:
        message_to_send (str): Message to send
        data_to_send (dict): data structure to go along with the
            message, defaults to empty dict.
    """
    data_to_send = data_to_send or {}

    url = MessageBusClient.build_url(config.get("host"),
                                     config.get("port"),
                                     config.get("route"),
                                     config.get("ssl"))

    # Send the provided message/data
    ws = create_connection(url)
    packet = Message(message_to_send, data_to_send).serialize()
    ws.send(packet)
    ws.close()



"""
Tools and constructs that are useful together with the messagebus.
"""

from .scheduler import EventScheduler
from .utils import create_echo_function

__all__ = [EventScheduler, create_echo_function]



from time import time
from unittest import TestCase

from alice_bus_client import Message
from alice_bus_client.message import dig_for_message


def get_message_standard(message):
    return dig_for_message()


def get_message_alt_name(msg):
    return dig_for_message()


def get_message_no_name(_):
    return dig_for_message()


class TestMessage(TestCase):
    def test_serialize_deserialize(self):
        """Assert that a serized message is recreated when deserialized."""
        source = Message('test_type',
                         data={'robot': 'marvin', 'android': 'data'},
                         context={'origin': 'earth'})
        msg_string = source.serialize()
        reassembled = Message.deserialize(msg_string)
        self.assertEqual(source.msg_type, reassembled.msg_type)
        self.assertEqual(source.data, reassembled.data)
        self.assertEqual(source.context, reassembled.context)

    def test_response(self):
        """Assert that the .response is added to the message type for response.
        """
        source = Message('test_type',
                         data={'robot': 'marvin', 'android': 'data'},
                         context={'origin': 'earth'})
        response_msg = source.response()
        self.assertEqual(response_msg.msg_type, "test_type.response")
        self.assertEqual(response_msg.data, {})
        self.assertEqual(response_msg.context, source.context)

    def test_reply(self):
        """Assert that the source and destination are swapped"""
        source = Message('test_type',
                         data={'robot': 'marvin', 'android': 'data'},
                         context={'source': 'earth',
                                  'destination': 'alpha centauri'})

        reply_msg = source.reply('reply_type')
        self.assertEqual(reply_msg.context["source"],
                         source.context["destination"])
        self.assertEqual(reply_msg.context["destination"],
                         source.context["source"])

        # assert that .response calls .reply internally as stated in docstrings
        response_msg = source.response()
        self.assertEqual(response_msg.context, reply_msg.context)

    def test_dig_for_message_simple(self):
        test_msg = Message("test message", {"test": "data"}, {"time": time()})
        self.assertEqual(test_msg, get_message_standard(test_msg))
        test_msg = Message("test message", {"test": "data"}, {"time": time()})
        self.assertEqual(test_msg, get_message_alt_name(test_msg))
        test_msg = Message("test message", {"test": "data"}, {"time": time()})
        self.assertEqual(test_msg, get_message_no_name(test_msg))

    def test_dig_for_message_nested(self):
        message = Message("test message", {"test": "data"}, {"time": time()})

        def simple_wrapper():
            return get_message_no_name(message)

        self.assertEqual(simple_wrapper(), message)

        message = Message("test message", {"test": "data"}, {"time": time()})

        def get_message():
            return dig_for_message()

        def wrapper_method(msg):
            self.assertEqual(msg, get_message())

        wrapper_method(message)

    def test_dig_for_message_invalid_type(self):
        # Message that should be ignored
        _ = Message("test message", {"test": "data"}, {"time": time()})

        def wrapper_method(_):
            return dig_for_message()
        self.assertIsNone(wrapper_method(dict()))

    def test_dig_for_message_no_method_call(self):
        # Message that should be ignored
        _ = Message("test message", {"test": "data"}, {"time": time()})
        self.assertIsNone(dig_for_message())



from .client.client import MessageBusClient
from .message import Message
from .send_func import send
from .conf import client_from_config

"""
Alice Messagebus Client.

A client for using the Alice message bus.

The messagebus client allows the developer to register handlers for
specific message types and emitting messages to other services and
clients connected to the bus.
"""

__all__ = [
    MessageBusClient,
    Message,
    send,
    client_from_config
]



from unittest.mock import Mock

from pyee import ExecutorEventEmitter

from alice_bus_client import MessageBusClient, Message
from alice_bus_client.client import MessageWaiter, MessageCollector


WS_CONF = {
    'websocket': {
        "host": "testhost",
        "port": 1337,
        "route": "/core",
        "ssl": False
    }
}


class TestMessageBusClient:
    def test_build_url(self):
        url = MessageBusClient.build_url('localhost', 1337, '/core', False)
        assert url == 'ws://localhost:1337/core'
        ssl_url = MessageBusClient.build_url('sslhost', 443, '/core', True)
        assert ssl_url == 'wss://sslhost:443/core'

    def test_create_client(self):
        mc = MessageBusClient()
        assert mc.client.url == 'ws://0.0.0.0:8181/core'

    def test_create_client_default_executor(self):
        mc = MessageBusClient()
        assert type(mc.emitter) == ExecutorEventEmitter

    def test_create_client_custom_executor(self):
        mock_emitter = Mock()
        mc = MessageBusClient(emitter=mock_emitter)
        assert mc.emitter == mock_emitter


class TestMessageWaiter:
    def test_message_wait_success(self):
        bus = Mock()
        waiter = MessageWaiter(bus, 'delayed.message')
        bus.once.assert_called_with('delayed.message', waiter._handler)

        test_msg = Mock(name='test_msg')
        waiter._handler(test_msg)  # Inject response

        assert waiter.wait() == test_msg

    def test_message_wait_timeout(self):
        bus = Mock()
        waiter = MessageWaiter(bus, 'delayed.message')
        bus.once.assert_called_with('delayed.message', waiter._handler)

        assert waiter.wait(0.3) is None


class TestMessageCollector:
    def test_message_wait_success(self):
        bus = Mock()
        collector = MessageCollector(bus, Message('delayed.message'),
                                     min_timeout=0.0, max_timeout=2.0)

        test_register = Mock(name='test_register')
        test_register.data = {
            'query': collector.collect_id,
            'timeout': 5,
            'handler': 'test_handler1'
        }
        collector._register_handler(test_register)  # Inject response

        test_response = Mock(name='test_register')
        test_response.data = {
            'query': collector.collect_id,
            'handler': 'test_handler1'
        }
        collector._receive_response(test_response)

        assert collector.collect() == [test_response]

    def test_message_drop_invalid(self):
        bus = Mock()
        collector = MessageCollector(bus, Message('delayed.message'),
                                     min_timeout=0.0, max_timeout=2.0)

        valid_register = Mock(name='valid_register')
        valid_register.data = {
            'query': collector.collect_id,
            'timeout': 5,
            'handler': 'test_handler1'
        }
        invalid_register = Mock(name='invalid_register')
        invalid_register.data = {
            'query': 'asdf',
            'timeout': 5,
            'handler': 'test_handler1'
        }
        collector._register_handler(valid_register)  # Inject response
        collector._register_handler(invalid_register)  # Inject response

        valid_response = Mock(name='valid_register')
        valid_response.data = {
            'query': collector.collect_id,
            'handler': 'test_handler1'
        }
        invalid_response = Mock(name='invalid_register')
        invalid_response.data = {
            'query': 'asdf',
            'handler': 'test_handler1'
        }
        collector._receive_response(valid_response)
        collector._receive_response(invalid_response)
        assert collector.collect() == [valid_response]



from threading import Event


class MessageWaiter:
    """Wait for a single message.

    Encapsulate the wait for a message logic separating the setup from
    the actual waiting act so the waiting can be setuo, actions can be
    performed and _then_ the message can be waited for.

    Argunments:
        bus: Bus to check for messages on
        message_type: message type to wait for
    """
    def __init__(self, bus, message_type):
        self.bus = bus
        self.msg_type = message_type
        self.received_msg = None
        # Setup response handler
        self.response_event = Event()
        self.bus.once(message_type, self._handler)

    def _handler(self, message):
        """Receive response data."""
        self.received_msg = message
        self.response_event.set()

    def wait(self, timeout=3.0):
        """Wait for message.

        Arguments:
            timeout (int or float): seconds to wait for message

        Returns:
            Message or None
        """
        self.response_event.wait(timeout)
        if not self.response_event.is_set():
            # Clean up the event handler
            try:
                self.bus.remove(self.msg_type, self._handler)
            except (ValueError, KeyError):
                # ValueError occurs on pyee 5.0.1 removing handlers
                # registered with once.
                # KeyError may theoretically occur if the event occurs as
                # the handler is removed
                pass
        return self.received_msg



import json

from .client import MessageBusClient

"""Functions for handling loading messagebus client configurations."""


def client_from_config(subconf='core', file_path='/etc/alice/bus.conf'):
    """Load messagebus configuration from file.

    The config is a basic json file with a number of "sub configurations"

    Ex:
    {
      "core": {
        "route": "/core",
        "port": "8181"
      }
      "gui": {
        "route": "/gui",
        "port": "8811"
      }
    }

    Arguments:
        subconf:    configuration to choose from the file, defaults to "core"
                    if omitted.
        file_path:  path to the config file, defaults to /etc/alice/bus.conf
                    if omitted.
    Returns:
        MessageBusClient instance based on the selected config.
    """
    with open(file_path) as f:
        conf = json.load(f)

    return MessageBusClient(**conf[subconf])
